@@ -36,6 +36,7 @@
 #include "gvfsjobcloseread.h"
 #include "gvfsjobread.h"
 #include "gvfsjobseekread.h"
+#include "gvfsjobseekwrite.h"
 #include "gvfsmonitor.h"
 #include "gvfsjobsetdisplayname.h"
 #include "gvfsjobopenforwrite.h"
@@ -158,28 +159,6 @@ items_in_folder_equal (gconstpointer a, gconstpointer b)
   return FALSE;
 }
 
-static char *
-get_full_item_id (MsgDriveItem *item)
-{
-  const char *drive_id = msg_drive_item_get_drive_id (item);
-
-  if (!drive_id)
-    drive_id = "";
-
-  return g_strconcat (drive_id, msg_drive_item_get_id (item), NULL);
-}
-
-static char *
-get_full_parent_id (MsgDriveItem *item)
-{
-  const char *drive_id = msg_drive_item_get_drive_id (item);
-
-  if (!drive_id)
-    drive_id = "";
-
-  return g_strconcat (drive_id, msg_drive_item_get_parent_id (item), NULL);
-}
-
 static void
 log_dir_items (GVfsBackendOnedrive *self)
 {
@@ -193,10 +172,8 @@ log_dir_items (GVfsBackendOnedrive *self)
   g_hash_table_iter_init (&iter, self->dir_items);
   while (g_hash_table_iter_next (&iter, (gpointer *) &key, (gpointer *) &item))
     {
-      g_autofree char *id = get_full_item_id (MSG_DRIVE_ITEM (item));
-
       g_debug ("  Real ID = %s, (%s, %s) -> %p, %d\n",
-               id,
+               msg_drive_item_get_id (MSG_DRIVE_ITEM (item)),
                key->name_or_id,
                key->parent_id,
                item,
@@ -348,8 +325,8 @@ insert_item (GVfsBackendOnedrive *self,
              MsgDriveItem        *item)
 {
   DirItemsKey *k;
-  g_autofree char *id = NULL;
-  g_autofree char *parent_id = NULL;
+  const char *id;
+  const char *parent_id;
   const char *name;
   gint64 *timestamp;
 
@@ -359,11 +336,11 @@ insert_item (GVfsBackendOnedrive *self,
   g_object_set_data_full (G_OBJECT (item), "timestamp", timestamp, g_free);
 
   /* Add item to items hash */
-  id = get_full_item_id (item);
+  id = msg_drive_item_get_id (item);
   g_hash_table_insert (self->items, g_strdup (id), g_object_ref (item));
 
   /* Add item to parent dir item hash */
-  parent_id = get_full_parent_id (item);
+  parent_id = msg_drive_item_get_parent_id (item);
   k = dir_items_key_new (id, parent_id);
   g_hash_table_insert (self->dir_items, k, g_object_ref (item));
   g_debug ("  insert_item: Inserted real     (%s, %s) -> %p\n", id, parent_id, item);
@@ -382,9 +359,10 @@ insert_custom_item (GVfsBackendOnedrive *self,
                     const char          *parent_id)
 {
   DirItemsKey *k;
-  g_autofree char *id = get_full_item_id (item);
+  const char *id;
   const char *name;
 
+  id = msg_drive_item_get_id (item);
   name = msg_drive_item_get_name (item);
 
   g_hash_table_insert (self->items, g_strdup (id), g_object_ref (item));
@@ -407,11 +385,11 @@ remove_item_full (GVfsBackendOnedrive *self,
                   MsgDriveItem        *item)
 {
   DirItemsKey *k;
-  g_autofree char *id = NULL;
-  g_autofree char *parent_id = NULL;
+  const char *id;
   const char *name;
+  const char *parent_id;
 
-  id = get_full_item_id (item);
+  id = msg_drive_item_get_id (item);
   name = msg_drive_item_get_name (item);
 
   /* Remove item from hash */
@@ -420,7 +398,7 @@ remove_item_full (GVfsBackendOnedrive *self,
   if (is_shared_with_me (item))
     g_hash_table_remove (self->dir_timestamps, SHARED_WITH_ME_ID);
 
-  parent_id = get_full_parent_id (item);
+  parent_id = msg_drive_item_get_parent_id (item);
   g_hash_table_remove (self->dir_timestamps, parent_id);
 
   k = dir_items_key_new (id, parent_id);
@@ -447,9 +425,9 @@ remove_dir (GVfsBackendOnedrive *self,
 {
   GHashTableIter iter;
   MsgDriveItem *item;
-  g_autofree char *parent_id = NULL;
+  const char *parent_id = NULL;
 
-  parent_id = get_full_item_id (parent);
+  parent_id = msg_drive_item_get_id (parent);
 
   g_hash_table_remove (self->dir_timestamps, parent_id);
 
@@ -457,10 +435,8 @@ remove_dir (GVfsBackendOnedrive *self,
   while (g_hash_table_iter_next (&iter, NULL, (gpointer *) &item))
     {
       DirItemsKey *k;
-      g_autofree char *id = NULL;
 
-      id = get_full_item_id (item);
-      k = dir_items_key_new (id, parent_id);
+      k = dir_items_key_new (msg_drive_item_get_id (item), parent_id);
       if (g_hash_table_lookup (self->dir_items, k) != NULL)
         {
           g_object_ref (item);
@@ -490,13 +466,11 @@ is_dir_listing_valid (GVfsBackendOnedrive *self,
                       MsgDriveItem        *parent)
 {
   gint64 *timestamp;
-  g_autofree char *id = NULL;
 
   if (parent == self->root)
     return TRUE;
 
-  id = get_full_item_id (parent);
-  timestamp = g_hash_table_lookup (self->dir_timestamps, id);
+  timestamp = g_hash_table_lookup (self->dir_timestamps, msg_drive_item_get_id (parent));
   if (timestamp != NULL)
     return (g_get_real_time () - *timestamp < REBUILD_ENTRIES_TIMEOUT * G_USEC_PER_SEC);
 
@@ -559,7 +533,7 @@ rebuild_dir (GVfsBackendOnedrive  *self,
 
   timestamp = g_new (gint64, 1);
   *timestamp = g_get_real_time ();
-  g_hash_table_insert (self->dir_timestamps, get_full_item_id (parent), timestamp);
+  g_hash_table_insert (self->dir_timestamps, g_strdup (msg_drive_item_get_id (parent)), timestamp);
 
   for (GList *l = items; l != NULL; l = l->next)
     {
@@ -581,10 +555,10 @@ resolve_child (GVfsBackendOnedrive  *self,
   GError *local_error = NULL;
   DirItemsKey *k;
   MsgDriveItem *item;
-  g_autofree char *parent_id = NULL;
+  const char *parent_id;
   gboolean is_shared_with_me_dir = (parent == self->shared_with_me_dir);
 
-  parent_id = get_full_item_id (parent);
+  parent_id = msg_drive_item_get_id (parent);
   k = dir_items_key_new (basename, parent_id);
 
   item = g_hash_table_lookup (self->dir_items, k);
@@ -657,8 +631,7 @@ resolve (GVfsBackendOnedrive  *self,
 
   if (out_path != NULL)
     {
-      g_autofree char *id = get_full_item_id (ret_val);
-      char *tmp  = g_build_path ("/", *out_path, id, NULL);
+      char *tmp  = g_build_path ("/", *out_path, msg_drive_item_get_id (ret_val), NULL);
       g_free (*out_path);
       *out_path = tmp;
     }
@@ -720,7 +693,7 @@ build_file_info (GVfsBackendOnedrive    *self,
 {
   g_autofree char *mime_type = NULL;
   GFileType file_type;
-  g_autofree char *id = NULL;
+  const char *id;
   const char *name;
   const char *user;
   const char *etag;
@@ -728,7 +701,7 @@ build_file_info (GVfsBackendOnedrive    *self,
   gboolean is_root = FALSE;
   gboolean is_home = (item == self->home);
   gboolean is_shared_with_me = (item == self->shared_with_me_dir);
-  gboolean uncertain_content_type = FALSE;
+  gboolean uncertain_content_type;
 
   if (MSG_IS_DRIVE_ITEM_FOLDER (item))
     is_folder = TRUE;
@@ -800,7 +773,7 @@ build_file_info (GVfsBackendOnedrive    *self,
 
   g_file_info_set_file_type (info, file_type);
 
-  id = get_full_item_id (item);
+  id = msg_drive_item_get_id (item);
   g_file_info_set_attribute_string (info, G_FILE_ATTRIBUTE_ID_FILE, id);
 
   if (is_root)
@@ -927,7 +900,7 @@ g_vfs_backend_onedrive_delete (GVfsBackend   *_self,
       goto out;
     }
 
-  id = get_full_item_id (item);
+  id = g_strdup (msg_drive_item_get_id (item));
 
   parent = resolve_dir (self, filename, cancellable, NULL, NULL, &error);
   if (error != NULL)
@@ -1039,7 +1012,7 @@ g_vfs_backend_onedrive_enumerate (GVfsBackend           *_self,
 
   g_vfs_job_succeeded (G_VFS_JOB (job));
 
-  id = get_full_item_id (item);
+  id = g_strdup (msg_drive_item_get_id (item));
   is_shared_with_me_dir = (item == self->shared_with_me_dir);
 
   g_hash_table_iter_init (&iter, self->items);
@@ -1048,7 +1021,7 @@ g_vfs_backend_onedrive_enumerate (GVfsBackend           *_self,
       DirItemsKey *k;
       g_autofree char *child_id = NULL;
 
-      child_id = get_full_item_id (child);
+      child_id = g_strdup (msg_drive_item_get_id (child));
       k = dir_items_key_new (child_id, id);
 
       if ((is_shared_with_me_dir && is_shared_with_me (child)) ||
@@ -1209,7 +1182,6 @@ g_vfs_backend_onedrive_mount (GVfsBackend  *_self,
       MsgDrive *drive = NULL;
       MsgDriveItem *item = NULL;
       g_autoptr (GError) local_error = NULL;
-      const char *drive_name = NULL;
 
       drive = MSG_DRIVE (l->data);
       item = msg_drive_service_get_root (self->service, drive, cancellable, &local_error);
@@ -1222,15 +1194,6 @@ g_vfs_backend_onedrive_mount (GVfsBackend  *_self,
       if (!self->home)
         {
           self->home = item;
-        }
-
-      drive_name = msg_drive_get_name (drive);
-      if (drive_name)
-        {
-          msg_drive_item_set_name (item, drive_name);
-        }
-      else
-        {
           msg_drive_item_set_name (item, _("My Files"));
         }
 
@@ -1539,6 +1502,32 @@ g_vfs_backend_onedrive_seek_on_read (GVfsBackend       *_self,
   g_debug ("- seek_on_read\n");
 }
 
+static void
+g_vfs_backend_onedrive_seek_on_write (GVfsBackend       *backend,
+                                      GVfsJobSeekWrite  *job,
+                                      GVfsBackendHandle  handle,
+                                      goffset            offset,
+                                      GSeekType          type)
+{
+  WriteHandle *wh = handle;
+  g_autoptr (GError) error = NULL;
+
+  g_debug ("+ seek_on_write: %p\n", handle);
+
+  if (g_seekable_seek (G_SEEKABLE (wh->stream), offset, job->seek_type, NULL, &error))
+    {
+      g_vfs_job_seek_write_set_offset (job, g_seekable_tell (G_SEEKABLE (wh->stream)));
+      g_vfs_job_succeeded (G_VFS_JOB (job));
+    }
+  else
+    {
+      g_warning ("Could not seek: %s", error->message);
+      g_vfs_job_failed_from_error (G_VFS_JOB (job), error);
+    }
+
+  g_debug ("- seek_on_write\n");
+}
+
 static void
 close_read_cb (GObject      *source_object,
                GAsyncResult *res,
@@ -1723,6 +1712,7 @@ g_vfs_backend_onedrive_create (GVfsBackend         *_self,
 
   handle = write_handle_new (new_item, stream, filename, item_path);
   g_vfs_job_open_for_write_set_handle (job, handle);
+  g_vfs_job_open_for_write_set_can_seek (job, TRUE);
   g_vfs_job_succeeded (G_VFS_JOB (job));
 
  out:
@@ -1906,18 +1896,11 @@ g_vfs_backend_onedrive_replace (GVfsBackend         *_self,
       insert_item (self, MSG_DRIVE_ITEM (new_item));
       g_hash_table_foreach (self->monitors, emit_create_event, item_path);
 
-      stream = msg_drive_service_update (self->service, new_item, cancellable, &error);
-      if (error != NULL)
-        {
-          g_vfs_job_failed_from_error (G_VFS_JOB (job), error);
-          g_error_free (error);
-          goto out;
-        }
-
-      handle = write_handle_new (MSG_DRIVE_ITEM (new_item), stream, filename, item_path);
+      handle = write_handle_new (MSG_DRIVE_ITEM (new_item), NULL, filename, item_path);
     }
 
   g_vfs_job_open_for_write_set_handle (job, handle);
+  g_vfs_job_open_for_write_set_can_seek (job, TRUE);
   g_vfs_job_succeeded (G_VFS_JOB (job));
 
 out:
@@ -1980,6 +1963,7 @@ g_vfs_backend_onedrive_class_init (GVfsBackendOnedriveClass *klass)
   backend_class->try_query_info_on_read = g_vfs_backend_onedrive_try_query_info_on_read;
   backend_class->try_query_info_on_write = g_vfs_backend_onedrive_try_query_info_on_write;
   backend_class->seek_on_read = g_vfs_backend_onedrive_seek_on_read;
+  backend_class->seek_on_write = g_vfs_backend_onedrive_seek_on_write;
   backend_class->set_display_name = g_vfs_backend_onedrive_set_display_name;
   backend_class->try_read = g_vfs_backend_onedrive_try_read;
   backend_class->replace = g_vfs_backend_onedrive_replace;
